import * as animation from "./animation.js";
import * as interpreter from "./interpreter.js";
import * as renderer from "./renderer.js";
import { DancerIdentity } from "./danceCommon.js";
import { LibFigureDance, LibFigureMove, Move } from "./libfigureMapper.js";
import { animateLowLevelMove, LowLevelMove } from "./lowLevelMove.js";
import { figureBeats, figureToHtml } from "./libfigure/define-figure.js";
import { labelForBeats } from "./libfigure/dance.js";

const body = document.querySelector('body')!;

const wrapperDiv = document.createElement('div');
const canvasDiv = document.createElement('div');
canvasDiv.id = 'canvasDiv';

const canvas = document.createElement('canvas');
for (const str of ["left", "right", "top", "bottom"]) {
  const hallLabel = document.createElement('span');
  hallLabel.innerText = str;
  hallLabel.classList.add(str);
  hallLabel.classList.add('hallLabel');
  canvasDiv.appendChild(hallLabel);
}
canvasDiv.appendChild(canvas);
wrapperDiv.appendChild(canvasDiv);
body.appendChild(wrapperDiv);

const beatSliderLabel = document.createElement('label');
beatSliderLabel.innerText = 'Beat: ';
const beatSlider = document.createElement('input');
beatSlider.type = 'range';
beatSlider.min = '0';
beatSlider.max = '64';
beatSlider.step = 'any';
beatSlider.value = '0';
beatSliderLabel.appendChild(beatSlider);

const beatDisplay = document.createElement('span');
beatDisplay.className = 'beatDisplay';
beatDisplay.innerText = '0.0';

const ctx = canvas.getContext('2d')!;

// TODO Redo this on resize?
const w = canvas.offsetWidth;
const h = canvas.offsetHeight;
const s = window.devicePixelRatio;

canvas.width = w * s;
canvas.height = h * s;

ctx.translate(canvas.width / 2.0, canvas.height / 2.0);
const scale = 10;
ctx.scale(canvas.height / scale, -canvas.height / scale);
const r = new renderer.Renderer(canvas, ctx);
r.extraLines = 3;
r.extraSets = 3;
r.clear();
r.animation = interpreter.interpretedAnimation;
beatSlider.max = r.animation.numBeats.toString();
r.drawSetsWithTrails(0);

const bpmSelector = document.createElement('input');
bpmSelector.type = 'number';
bpmSelector.value = '180';
bpmSelector.id = 'bpm';
bpmSelector.style.width = '4em';
const bpmLabel = document.createElement('label');
bpmLabel.innerText = 'BPM';
bpmLabel.htmlFor = 'bpm';

const progressionSelector = document.createElement('input');
progressionSelector.type = 'number';
progressionSelector.value = '0';
progressionSelector.id = 'progression';
progressionSelector.style.width = '4em';
const progressionLabel = document.createElement('label');
progressionLabel.innerText = 'Progression';
progressionLabel.htmlFor = 'progression';

const playButton = document.createElement('button');
playButton.innerText = "Play";

const movesList = document.createElement('ul');

wrapperDiv.appendChild(movesList);
wrapperDiv.appendChild(document.createElement('br'));
wrapperDiv.appendChild(bpmSelector);
wrapperDiv.appendChild(bpmLabel);
wrapperDiv.appendChild(playButton);
wrapperDiv.appendChild(document.createElement('br'));
wrapperDiv.appendChild(progressionSelector);
wrapperDiv.appendChild(progressionLabel);
wrapperDiv.appendChild(document.createElement('br'));
wrapperDiv.appendChild(beatSliderLabel);
wrapperDiv.appendChild(beatDisplay);

function drawAtCurrentBeat() {
  r.drawSetsWithTrails(beatSlider.valueAsNumber, progressionSelector.valueAsNumber);

  const currentMoves = movesList.getElementsByClassName('currentMove');
  for (let i = 0; i < currentMoves.length; i++) {
    currentMoves[i].classList.remove('currentMove');
  }
  movesList.getElementsByClassName('moveForBeat_' + Math.floor(beatSlider.valueAsNumber))[0]
    .classList.add('currentMove');
}

beatSlider.addEventListener('input', (ev) => {
  drawAtCurrentBeat();
  beatDisplay.innerText = beatSlider.valueAsNumber.toFixed(1);
  restartAnimation(false);
});

progressionSelector.addEventListener('input', (ev) => {
  drawAtCurrentBeat();
  restartAnimation(false);
});

let cancelAnim : number | undefined = undefined;
function restartAnimation(startIfPaused: boolean) {
  if (cancelAnim !== undefined) {
    cancelAnimationFrame(cancelAnim);
  } else if (!startIfPaused) {
    return;
  }
  const bpm = parseFloat(bpmSelector.value);
  if (bpm < 0) {
    playAnimation(
      bpmSelector.valueAsNumber,
      beatSlider.value === '0' ? r.animation.numBeats : beatSlider.valueAsNumber,
      0);
  } else if (bpm > 0) {
    playAnimation(
      bpmSelector.valueAsNumber,
      beatSlider.value === beatSlider.max ? 0 : beatSlider.valueAsNumber,
      r.animation.numBeats);
  }
}
playButton.addEventListener('click', (ev) => {
  if (cancelAnim !== undefined) {
    cancelAnimationFrame(cancelAnim);
    cancelAnim = undefined;
    playButton.innerText = 'Play';
  } else {
    restartAnimation(true);
  }
});
bpmSelector.addEventListener('change', (ev) => {
  restartAnimation(false);
});

// Two Hearts in Time by Isaac Banner. Selected arbitrarily.
const exampleDance: LibFigureDance = [{ "parameter_values": [true, 8], "move": "petronella" }, { "parameter_values": [true, 8], "move": "petronella" }, { "parameter_values": ["neighbors", "balance", 16], "move": "swing" }, { "parameter_values": ["ladles", true, 540, 8], "move": "allemande" }, { "parameter_values": ["partners", "none", 8], "move": "swing" }, { "parameter_values": ["gentlespoons", 360, 6], "move": "mad robin" }, { "parameter_values": [true, 270, 6], "move": "circle" }, { "parameter_values": ["partners", 4], "move": "California twirl", "progression": 1 }];

const dialect = {
  moves: {},
  dancers: {
    ladle: "robin",
    ladles: "robins",
    gentlespoon: "lark",
    gentlespoons: "larks",
    "first ladle": "first robin",
    "second ladle": "second robin",
    "first gentlespoon": "first lark",
    "second gentlespoon": "second lark",
  },
};

let dance: LibFigureDance;
function buildMovesList() {
  removeAllChildNodes(movesList);

  let currentBeat: number = 0;
  let lastItem: HTMLLIElement | undefined;
  for (const figure of dance) {
    const startBeat = currentBeat;
    currentBeat += figureBeats(figure);
    const moveItem = document.createElement('li');
    moveItem.innerHTML = "[" + labelForBeats(Math.floor(startBeat / 16) * 16) + " "
      + "(" + startBeat + "-" + currentBeat + ")] "
      + figureToHtml(figure, dialect);
    moveItem.classList.add('move');
    for (let beat = startBeat; beat < currentBeat; beat++) {
      moveItem.classList.add('moveForBeat_' + beat);
    }
    if (startBeat === 0) moveItem.classList.add('currentMove');
    moveItem.addEventListener('click', (ev) => {
      setBeat(startBeat);
      restartAnimation(false);
    });
    lastItem = moveItem;
    movesList.appendChild(moveItem);
  }
  if (lastItem) {
    lastItem.classList.add('moveForBeat_' + currentBeat);
  }
}

function setBeat(beat: number) {
  beatSlider.value = beat.toString();
  beatDisplay.innerText = beat.toFixed(1);

  drawAtCurrentBeat();
}
function playAnimation(bpm: number, start: number, end: number) {
  const startTime = Date.now();
  const msPerBeat = (60 * 1000) / bpm;

  function anim() {
    const now = Date.now();
    const msElapsed = now - startTime;

    let beat = start + msElapsed / msPerBeat;
    let changedProgression = false;
    if (bpm > 0 && beat > end) {
      beat -= r.animation.numBeats;
      progressionSelector.valueAsNumber++;
      changedProgression = true;
    }
    if (bpm < 0 && beat < end) {
      beat += r.animation.numBeats;
      progressionSelector.valueAsNumber--;
      changedProgression = true;
    }

    setBeat(beat);
    if (changedProgression) {
      restartAnimation(true);
      return;
    }
    cancelAnim = requestAnimationFrame(anim);
    playButton.innerText = 'Pause';
  }
  anim();
}

const danceJsonArea = document.createElement('textarea');
danceJsonArea.value = JSON.stringify(exampleDance, undefined, 2);
danceJsonArea.rows = 15;
danceJsonArea.cols = 30;
const loadDanceButton = document.createElement('button');
loadDanceButton.innerText = 'Load Dance';
wrapperDiv.appendChild(document.createElement('br'));
wrapperDiv.appendChild(danceJsonArea);
wrapperDiv.appendChild(loadDanceButton);

const table = document.createElement('table');
function loadDance() {
  dance = JSON.parse(danceJsonArea.value);
  r.animation = interpreter.loadDance(dance);
  if (cancelAnim !== undefined) {
    cancelAnimationFrame(cancelAnim);
    playButton.innerText = 'Play';
  }
  beatSlider.value = '0';
  beatSlider.max = r.animation.numBeats.toString();
  beatDisplay.innerText = '0.0';
  r.drawSetsWithTrails(0);
  buildDebugTable();
  buildMovesList();
}
loadDanceButton.addEventListener('click', loadDance);
loadDance();

function createJsonCell(content: any, rowSpan?: number, id?: DancerIdentity) {
  const cell = document.createElement('td');
  const pre = document.createElement('pre');

  // from https://stackoverflow.com/a/56150320
  function replacer(key, value) {
    if (value instanceof Map) {
      let res = {};
      for (var [k, v] of value.entries()) {
        res[k === undefined ? "undefined" : k.toString()] = v;
      }
      return res;
    } else if (!value) {
      return value;
    } else if (value.enumValue) {
      return value.enumValue;
    } else {
      return value;
    }
  }

  pre.innerText = content ? JSON.stringify(content, replacer, 2) : "";
  cell.appendChild(pre);
  if (rowSpan) {
    cell.rowSpan = rowSpan;
  }
  if (id) {
    cell.classList.add(id.coupleRole.toString());
    cell.classList.add(id.danceRole.toString());
  }
  return cell;
}
function createHeaderCell(content: string, id?: DancerIdentity) {
  const cell = document.createElement('th');
  cell.innerText = content;
  if (id) {
    cell.classList.add(id.coupleRole.toString());
    cell.classList.add(id.danceRole.toString());
    cell.colSpan = 2;
  }
  return cell;
}

const showDebug = document.createElement('input');
showDebug.type = 'checkbox';
showDebug.id = 'showDebug';
const showDebugLabel = document.createElement('label');
showDebugLabel.innerText = 'Show Debug Information';
showDebugLabel.htmlFor = 'showDebug';

body.appendChild(showDebug);
body.appendChild(showDebugLabel);

table.id = 'debug'

function removeAllChildNodes(html: HTMLElement) {
  while (html.childNodes.length > 0) {
    html.removeChild(html.childNodes[html.childNodes.length - 1]);
  }
}
function buildDebugTable() {
  removeAllChildNodes(table);

  const headerRow = document.createElement('tr');
  const roles = [DancerIdentity.OnesLark, DancerIdentity.OnesRobin,
  DancerIdentity.TwosLark, DancerIdentity.TwosRobin];
  headerRow.appendChild(createHeaderCell("Move"));
  for (const role of roles) {
    headerRow.appendChild(createHeaderCell(role.toString(), role));
  }
  table.appendChild(headerRow);

  const byMove: { move: Move, byRole: Map<DancerIdentity, { lowLevelMove?: LowLevelMove, animationSegment: animation.AnimationSegment, numSegments?: number }[]> }[] = [];
  for (const [role, moveList] of interpreter.interpretedDance.entries()) {
    for (const move of moveList) {
      let entry = byMove.find(el => el.move === move.move);
      if (!entry) {
        entry = { move: move.move, byRole: new Map<DancerIdentity, { lowLevelMove?: LowLevelMove, animationSegment: animation.AnimationSegment, numSegments?: number }[]>() };
        byMove.push(entry);
      }
      let forRole = entry.byRole.get(role);
      if (!forRole) {
        forRole = [];
        entry.byRole.set(role, forRole);
      }
      const animation = animateLowLevelMove(move);
      let first = true;
      for (const animationSegment of animation) {
        if (first) {
          forRole.push({ lowLevelMove: move, animationSegment, numSegments: animation.length });
          first = false;
        } else {
          forRole.push({ animationSegment });
        }
      }
    }
  }

  for (const infoForMove of byMove) {
    const moveRow = document.createElement('tr');
    const numRows = Math.max(...[...infoForMove.byRole.values()].map(l => l.length));
    moveRow.appendChild(createJsonCell(infoForMove.move, numRows));

    for (let i = 0; i < numRows; i++) {
      const row = i == 0 ? moveRow : document.createElement('tr');

      for (const role of roles) {
        const lowLevelMove = infoForMove.byRole.get(role)?.at(i);
        if (!lowLevelMove) {
          row.appendChild(createJsonCell(undefined, undefined, role));
        } else if (lowLevelMove.lowLevelMove) {
          row.appendChild(createJsonCell(lowLevelMove.lowLevelMove, lowLevelMove.numSegments, role));
        }
        row.appendChild(createJsonCell(lowLevelMove?.animationSegment, undefined, role));
      }

      table.appendChild(row);
    }
  }
}
buildDebugTable();

body.appendChild(table);
